import importlib
import os
from pathlib import Path
from typing import Any, Optional

# Sentinel value for undefined argument
UNDEFINED = object()


def required_env(name: str, default: Any = UNDEFINED) -> str:
    """Load required environment variable.

    Args:
        name: Name of environment variable
        default: Value to use if variable is undefined.
                    If not given and variable is undefined, raises KeyError.

    """
    val = os.getenv(name, default)
    if val is UNDEFINED:
        raise KeyError(f"Missing required environment variable: {name}")
    return val


def import_by_name(name: str, caller: Optional[str] = None) -> Any:
    """Import module (or attribute) by name.

    Args:
        name: Import path, e.g. RPA.Robocorp.WorkItems.RobocorpAdapter
        caller: Caller file name

    """
    name = str(name)

    # Attempt import as path module
    try:
        return importlib.import_module(name)
    except ImportError:
        pass

    # Attempt import from calling file
    if caller is not None:
        try:
            module = importlib.import_module(caller)
            return getattr(module, name)
        except AttributeError:
            pass

    # Attempt import as path to attribute inside module
    if "." in name:
        try:
            path, attr = name.rsplit(".", 1)
            module = importlib.import_module(path)
            return getattr(module, attr)
        except (AttributeError, ImportError):
            pass

    raise ValueError(f"No module/attribute with name: {name}")


def resolve_path(path: str) -> Path:
    """Resolve a string-based path, and replace variables."""
    return Path(path).expanduser().resolve()


def url_join(*parts):
    """Join parts of URL and handle missing/duplicate slashes."""
    return "/".join(str(part).strip("/") for part in parts)



import logging
import os

from ._adapters import BaseSecretManager, FileSecrets, RobocorpVault
from ._secrets import SecretContainer
from ._utils import import_by_name, required_env


class Vault:
    def __init__(self, *args, **kwargs):
        """The selected adapter can be set with the environment variable.

        Use the environment variable ``RC_VAULT_SECRET_MANAGER``, or the
        argument ``default_adapter``.
        Defaults to Robocorp Vault if not defined.

        All other library arguments are passed to the adapter.

        Args:
            default_adapter: Override default secret adapter

        """
        self.logger = logging.getLogger(__name__)

        default = kwargs.pop("default_adapter", RobocorpVault)
        if "RC_VAULT_SECRET_MANAGER" in os.environ:
            adapter = required_env("RC_VAULT_SECRET_MANAGER", default)
        elif "RPA_SECRET_MANAGER" in os.environ:  # Backward-compatibility
            adapter = required_env("RPA_SECRET_MANAGER", default)
        else:
            adapter = default

        self._adapter_factory = self._create_factory(adapter, args, kwargs)
        self._adapter = None

    @property
    def adapter(self) -> BaseSecretManager:
        if self._adapter is None:
            self._adapter = self._adapter_factory()

        return self._adapter

    def _create_factory(self, adapter, args, kwargs):
        if isinstance(adapter, str):
            if adapter in ("FileSecrets", "RPA.Robocorp.Vault.FileSecrets"):
                adapter = FileSecrets
            else:
                try:
                    adapter = import_by_name(adapter, __name__)
                except Exception as e:
                    raise ValueError(
                        f"RC_VAULT_SECRET_MANAGER environment variable seems to "
                        f"have a wrong value: {adapter!r}.\n"
                        f'It should be unset to load from Control Room, "FileSecrets" '
                        f'to load from a file or the name of the "BaseSecretManager" '
                        f"implementation class."
                    ) from e

        if not issubclass(adapter, BaseSecretManager):
            raise ValueError(
                f"Adapter '{adapter}' does not inherit from BaseSecretManager"
            )

        def factory():
            return adapter(*args, **kwargs)

        return factory

    def get_secret(self, secret_name: str) -> SecretContainer:
        """Read a secret from the configured source, e.g. Robocorp Vault.

        Args:
            secret_name: Name of secret.

        Returns:
            `SecretContainer` object.
        """
        return self.adapter.get_secret(secret_name)

    def set_secret(self, secret: SecretContainer) -> None:
        """Create secret or overwrite existing.

        Args:
            secret: `SecretContainer` object, from e.g. `get_secret`.
        """
        self.adapter.set_secret(secret)



from contextlib import nullcontext
from functools import lru_cache
from typing import Any

from ._errors import RobocorpVaultError
from ._secrets import SecretContainer

__version__ = "1.3.3"
version_info = [int(x) for x in __version__.split(".")]


@lru_cache
def _get_vault():
    from ._vault import Vault

    return Vault()


def get_secret(name: str, hide: bool = True) -> SecretContainer:
    """Get a secret with the given name.

    Args:
        name: Name of secret to fetch
        hide: Hide secret values from log output

    Note:
        If `robocorp.log` is available in the environment, the `hide` argument
        controls if the given values are automatically hidden in the log
        output.

    Returns:
        Secret container of name, description, and key-value pairs

    Raises:
        RobocorpVaultError: Error with API request or response payload.
    """
    with _suppress_variables():
        vault = _get_vault()
        secret = vault.get_secret(name)

        if hide:
            _hide_secret_values(secret)

        return secret


def set_secret(secret: SecretContainer, hide: bool = True) -> None:
    """Set a secret value using an existing container.

    **Note:** If the secret already exists, all contents are replaced.

    Args:
        secret: Secret container, created manually or returned by `get_secret`
        hide: Hide secret values from log output

    Note:
        If `robocorp.log` is available in the environment, the `hide` argument
        controls if the given values are automatically hidden in the log
        output.

    Raises:
        RobocorpVaultError: Error with API request or response payload
    """
    with _suppress_variables():
        vault = _get_vault()
        vault.set_secret(secret)

        if hide:
            _hide_secret_values(secret)


def create_secret(
    name: str,
    values: dict[str, Any],
    description: str = "",
    exist_ok: bool = False,
    hide: bool = True,
) -> SecretContainer:
    """Create a new secret, or overwrite an existing one.

    Args:
        name: Name of secret
        values: Mapping of secret keys and values
        description: Optional description for secret
        exist_ok: Overwrite existing secret
        hide: Hide secret values from log output

    Note:
        If `robocorp.log` is available in the environment, the `hide` argument
        controls if the given values are automatically hidden in the log
        output.

    Returns:
        Secret container of name, description, and key-value pairs

    Raises:
        RobocorpVaultError: Error with API request or response payload
    """
    with _suppress_variables():
        vault = _get_vault()
        if not exist_ok:
            try:
                vault.get_secret(name)
            except Exception:
                pass
            else:
                raise RobocorpVaultError(f"Secret with name '{name}' already exists")

        secret = SecretContainer(
            name=name,
            description=description,
            values=values,
        )
        set_secret(secret, hide=hide)
        return secret


def _suppress_variables():
    try:
        from robocorp import log  # type: ignore [attr-defined]

        return log.suppress_variables()
    except ImportError:
        return nullcontext()


def _hide_secret_values(secret):
    try:
        from robocorp import log
    except ImportError:
        return

    for value in secret.values():
        s = str(value)
        log.hide_from_output(s)

        # Now, also take care of the case where the user does a repr(value)
        # and not just str(value) as in some places it's the repr(value)
        # that'll appear in the logs.
        r = repr(value)
        if r.startswith("'") and r.endswith("'"):
            r = r[1:-1]
        if r != s:
            log.hide_from_output(r)
            log.hide_from_output(r.replace("\\", "\\\\"))


__all__ = [
    "get_secret",
    "set_secret",
    "create_secret",
    "SecretContainer",
    "RobocorpVaultError",
]
